package de.lluni.javann;

import de.lluni.javann.layers.FCLayer;
import de.lluni.javann.layers.Layer;
import org.ejml.simple.SimpleMatrix;

import java.util.ArrayList;
import java.util.function.BiFunction;

public class Network {
    private ArrayList<Layer> layers;
    private BiFunction<SimpleMatrix, SimpleMatrix, Double> loss;
    private BiFunction<SimpleMatrix, SimpleMatrix, SimpleMatrix> lossPrime;

    public Network() {
        layers = new ArrayList<>();
    }

    public void addLayer(Layer layer) {
        layers.add(layer);
    }

    /**
     * Adds n neurons to a specific layer and also updates this and the next layer's weights and biases
     * @param layer index of layer in the ArrayList layers
     * @param n amount how many new neurons should be added
     */
    public void addNeuron(int layer, int n) {
        if (!(this.layers.get(layer) instanceof FCLayer)) {
            System.out.println("This layer is not a FCLayer");
        }

        ((FCLayer) this.layers.get(layer)).addNeuron(n);

        for (int i = layer + 1; i < this.layers.size(); i++) {
            if (this.getLayers().get(i) instanceof FCLayer) {
                ((FCLayer) this.layers.get(i)).updateInputSize(n);
                break;
            }
        }
    }

    /**
     * Sets loss function used to evaluate the network
     * Chosen from {@link de.lluni.javann.functions.LossFunctions}
     * @param loss loss function
     * @param lossPrime derivative of the loss function
     */
    public void use(BiFunction<SimpleMatrix, SimpleMatrix, Double> loss, BiFunction<SimpleMatrix, SimpleMatrix, SimpleMatrix> lossPrime) {
        this.loss = loss;
        this.lossPrime = lossPrime;
    }

    /**
     * Predicts labels corresponding to the given inputs
     * @param inputs inputs
     * @return array of predictions
     */
    public SimpleMatrix[] predict(SimpleMatrix[] inputs) {
        SimpleMatrix[] result = new SimpleMatrix[inputs.length];
        SimpleMatrix output;
        int i = 0;

        for (SimpleMatrix input : inputs) {
            output = input;
            for (Layer l : layers) {
                output = l.forwardPropagation(output);
            }
            result[i] = output;
            i++;
        }

        return result;
    }

    /**
     * Trains the network on a set of training data
     * @param X_train inputs
     * @param y_train labels
     * @param epochs amount of training iterations
     * @param learningRate step size of gradient descent
     */
    public void fit(SimpleMatrix[] X_train, SimpleMatrix[] y_train, int epochs, double learningRate) {
        int samples = X_train.length;

        for (int i = 0; i < epochs; i++) {
            double err = 0;
            for (int j = 0; j < samples; j++) {
                // forward propagation
                SimpleMatrix output = X_train[j];
                for (Layer l : layers) {
                    output = l.forwardPropagation(output);
                }

                // compute loss (for display purpose only)
                err = loss.apply(y_train[j], output);

                // backward propagation
                SimpleMatrix error = lossPrime.apply(y_train[j], output);
                for (int k = layers.size() - 1; k >= 0; k--) {
                    error = layers.get(k).backwardPropagation(error, learningRate / (i+1));
                }
            }
            // calculate average error on all samples
            err /= samples;
            System.out.println("epoch " + (i+1) + "/" + epochs + "  error=" + err);
        }
    }

    public ArrayList<Layer> getLayers() {
        return layers;
    }

    public void setLayers(ArrayList<Layer> layers) {
        this.layers = layers;
    }

    public BiFunction<SimpleMatrix, SimpleMatrix, Double> getLoss() {
        return loss;
    }

    public void setLoss(BiFunction<SimpleMatrix, SimpleMatrix, Double> loss) {
        this.loss = loss;
    }

    public BiFunction<SimpleMatrix, SimpleMatrix, SimpleMatrix> getLossPrime() {
        return lossPrime;
    }

    public void setLossPrime(BiFunction<SimpleMatrix, SimpleMatrix, SimpleMatrix> lossPrime) {
        this.lossPrime = lossPrime;
    }
}